import asyncio
import inspect
import logging
from functools import wraps
from os import getenv
from pathlib import Path

import click
import discord
import hupper
import requests

from spellbot._version import __version__
from spellbot.assets import ASSET_FILES, s
from spellbot.data import Data, Queue, User

# Application Paths
RUNTIME_ROOT = Path(".")
SCRIPTS_DIR = RUNTIME_ROOT / "scripts"
DB_DIR = RUNTIME_ROOT / "db"
DEFAULT_DB_URL = f"sqlite:///{DB_DIR}/spellbot.db"
TMP_DIR = RUNTIME_ROOT / "tmp"
MIGRATIONS_DIR = SCRIPTS_DIR / "migrations"

# Application Settings
ADMIN_ROLE = "SpellBot Admin"
CREATE_ENDPOINT = "https://us-central1-magic-night-30324.cloudfunctions.net/createGame"


class Direct:
    """Use this to indicate that a command should send a DM."""

    def __init__(self, response):
        self.response = response


def discord_user_from_name(channel, name):
    """Returns the discord user from the given channel and name."""
    if name is None:
        return None
    lname = name.lower()
    if hasattr(channel, "members"):  # channels
        members = channel.members
        return next(filter(lambda member: lname in str(member).lower(), members), None)
    else:  # direct messages
        recipient = channel.recipient
        rvalue = recipient if name in str(recipient).lower() else None
        return rvalue


def discord_user_from_id(channel, user_id):
    """Returns the discord user from the given channel and user id."""
    if user_id is None:
        return None
    iid = int(user_id)
    if hasattr(channel, "members"):  # channels
        members = channel.members
        return next(filter(lambda member: iid == member.id, members), None)
    else:  # direct messages
        recipient = channel.recipient
        return recipient if recipient.id == user_id else None


def discord_user_name(channel, name_or_id):
    """Returns the discord user name from the given channel and name or id."""
    if not name_or_id:
        return None
    user = (
        discord_user_from_id(channel, name_or_id)
        if isinstance(name_or_id, int) or name_or_id.isdigit()
        else discord_user_from_name(channel, name_or_id)
    )
    return str(user) if user else None


def discord_user_id(channel, name):
    """Returns the discord user id name from the given channel and name."""
    if not name:
        return None
    user = discord_user_from_name(channel, name)
    return user.id if user else None


def is_admin(channel, user_or_member):
    """Checks to see if given user or member has the admin role on this server."""
    member = (
        user_or_member
        if hasattr(user_or_member, "roles")  # members have a roles property
        else channel.guild.get_member(user_or_member.id)  # but users don't
    )
    return any(role.name == ADMIN_ROLE for role in member.roles) if member else False


def ensure_application_directories_exist():
    """Idempotent function to make sure needed application directories are there."""
    TMP_DIR.mkdir(exist_ok=True)
    DB_DIR.mkdir(exist_ok=True)


def command(allow_dm=True):
    """Decorator for bot command methods."""

    def callable(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            return func(*args, **kwargs)

        wrapped.is_command = True
        wrapped.allow_dm = allow_dm
        return wrapped

    return callable


class SpellBot(discord.Client):
    """Discord SpellTable Bot"""

    def __init__(
        self,
        token="",
        auth="",
        db_url=DEFAULT_DB_URL,
        log_level=logging.ERROR,
        metrics=False,
    ):
        logging.basicConfig(level=log_level)
        loop = asyncio.get_event_loop()
        super().__init__(loop=loop)
        self.token = token
        self.auth = auth

        # During the processing of a command there will be valid SQLAlchemy session
        # object available for use, commits and rollbacks are handled automatically.
        self.session = None

        # We have to make sure that DB_DIR exists before we try to create
        # the database as part of instantiating the Data object.
        ensure_application_directories_exist()
        self.data = Data(db_url)

        # build a list of commands supported by this bot by fetching @command methods
        members = inspect.getmembers(self, predicate=inspect.ismethod)
        self._commands = [
            member[0]
            for member in members
            if hasattr(member[1], "is_command") and member[1].is_command
        ]

    def run(self):  # pragma: no cover
        super().run(self.token)

    def create_game(self):
        headers = {"user-agent": f"spellbot/{__version__}", "key": self.auth}
        r = requests.post(CREATE_ENDPOINT, headers=headers)
        return r.json()["gameUrl"]

    def ensure_user_exists(self, user):
        """Ensures that the user row exists for the given discord user."""
        db_user = self.session.query(User).get(user.id)
        if not db_user:
            db_user = User(id=user.id)
            self.session.add(db_user)
        return db_user

    def paginate(self, text):
        """Discord responses must be 2000 characters of less; paginate breaks them up."""
        breakpoints = ["\n", ".", ",", "-"]
        remaining = text
        while len(remaining) > 2000:
            breakpoint = 1999

            for char in breakpoints:
                index = remaining.rfind(char, 1800, 1999)
                if index != -1:
                    breakpoint = index
                    break

            message = remaining[0 : breakpoint + 1]
            yield message
            remaining = remaining[breakpoint + 1 :]
            last_line_end = message.rfind("\n")
            if last_line_end != -1 and len(message) > last_line_end + 1:
                last_line_start = last_line_end + 1
            else:
                last_line_start = 0
            if message[last_line_start] == ">":
                remaining = f"> {remaining}"

        yield remaining

    @property
    def commands(self):
        """Returns a list of commands supported by this bot."""
        return self._commands

    async def process(self, message):
        """Process a command message."""
        tokens = message.content.split(" ")
        request, params = tokens[0].lstrip("!").lower(), tokens[1:]
        params = list(filter(None, params))  # ignore any empty string parameters
        if not request:
            return
        matching = [command for command in self.commands if command.startswith(request)]
        if not matching:
            await message.channel.send(s("not_a_command", request=request), file=None)
            return
        if len(matching) > 1 and request not in matching:
            possible = ", ".join(f"!{m}" for m in matching)
            await message.channel.send(s("did_you_mean", possible=possible), file=None)
        else:
            command = request if request in matching else matching[0]
            method = getattr(self, command)
            if not method.allow_dm and message.channel.type == "private":
                return
            logging.debug("%s (author=%s, params=%s)", command, message.author, params)
            # FIXME: Delete debug code.
            # async with message.channel.typing():
            ######################################
            self.session = self.data.Session()
            try:
                rvalue = await method(message.channel, message.author, params)
                self.session.commit()
            except:
                self.session.rollback()
                raise
            finally:
                self.session.close()
            ######################################
            if not rvalue:
                return
            response, attachment = rvalue
            if isinstance(response, Direct):
                send_direct = True
                response = response.response
            else:
                send_direct = False
            if not isinstance(response, list):
                response = [response]
            last_reply_index = len(response) - 1
            for n, reply in enumerate(response):
                if isinstance(reply, str):
                    pages = list(self.paginate(reply))
                    last_page_index = len(pages) - 1
                    for i, page in enumerate(pages):
                        file = (
                            attachment
                            if attachment is not None
                            and i == last_page_index
                            and n == last_reply_index
                            else None
                        )
                        if send_direct:
                            await message.author.send(page, file=file)
                        else:
                            await message.channel.send(page, file=file)
                elif isinstance(reply, discord.embeds.Embed):
                    file = (
                        attachment
                        if attachment is not None and n == last_reply_index
                        else None
                    )
                    if send_direct:
                        await message.author.send(embed=reply, file=file)
                    else:
                        await message.channel.send(embed=reply, file=file)
                else:
                    raise RuntimeError("non-string non-embed reply not supported")

    ##############################
    # Discord Client Behavior
    ##############################

    async def on_message(self, message):
        """Behavior when the client gets a message from Discord."""
        # don't respond to any bots
        if message.author.bot:
            return

        private = str(message.channel.type) == "private"

        # only respond in text channels and to direct messages
        if not private and str(message.channel.type) != "text":
            return

        # don't respond to yourself
        if message.author.id == self.user.id:
            return

        # only respond to command-like messages
        if not message.content.startswith("!"):
            return

        if not private:
            # check for admin authorized channels on this server
            guild_id = message.channel.guild.id
            rows = self.data.conn.execute(
                f"SELECT * FROM authorized_channels WHERE guild = {guild_id};"
            )
            authorized_channels = set(row["name"] for row in rows)
            if authorized_channels and message.channel.name not in authorized_channels:
                return

        await self.process(message)

    async def on_ready(self):
        """Behavior when the client has successfully connected to Discord."""
        logging.debug("logged in as %s", self.user)

    ##############################
    # Bot Command Functions
    ##############################

    # Any method of this class with a name that is decorated by @command is detected as a
    # bot command. These methods should have a signature like:
    #
    #     @command(allow_dm=True)
    #     def command_name(self, channel, author, params)
    #
    # - `channel` is the Discord channel where the command message was sent.
    # - `author` is the Discord author who sent the command.
    # - `params` are any space delimitered parameters also sent with the command.
    #
    # The return value for a command method can be `(string, discord.File)` where the
    # string is the response message the bot should send to the channel and the file
    # object is an attachment to send with the message. For no attachment, use `None`.
    #
    # You can also return `(embed, discord.File)` to respond with an embed (plus the
    # optional attachment as well if so desired).
    #
    # And finally you can also respond with a list if you want the bot to make multiple
    # replies. This works with both embeds and strings. For example:
    #
    #     return ["send", "multiple", "replies"], None
    #
    # Would trigger the bot to send three messages to the channel with no attachment.
    #
    # The docstring used for the command method will be automatically used as the help
    # message for the command. To document commands with parameters use a @ to delimit
    # the help message from the parameter documentation. For example:
    #
    #     """This is the help message for your command. @ [and] [these] [are] [params]"""
    #
    # A [parameter] is optional whereas a <parameter> is required.

    @command(allow_dm=True)
    async def help(self, channel, author, params):
        """
        Shows this help screen.
        """
        usage = "__**SpellBot Usage**__\n"
        for command in self.commands:
            method = getattr(self, command)
            doc = method.__doc__.split("@")
            use, params = doc[0], ", ".join([param.strip() for param in doc[1:]])
            use = inspect.cleandoc(use)
            use = use.replace("\n", " ")

            title = f"**!{command}**"
            if params:
                title = f"{title} _{params}_"
            usage += f"\n{title}"
            usage += f"\n>  {use}"
            usage += "\n"
        usage += "\n_SpellBot created by amy@lexicalunit.com_"
        return Direct(usage), None

    @command(allow_dm=True)
    async def hello(self, channel, author, params):
        """
        Says hello.
        """
        return "Hello!", None

    @command(allow_dm=True)
    async def about(self, channel, author, params):
        """
        Get information about SpellBot.
        """
        embed = discord.Embed(title="SpellBot")
        thumb = (
            "https://raw.githubusercontent.com/lexicalunit/spellbot/master/spellbot.png"
        )
        embed.set_thumbnail(url=thumb)
        version = f"[{__version__}](https://pypi.org/project/spellbot/{__version__}/)"
        embed.add_field(name="Version", value=version)
        embed.add_field(
            name="Package", value="[PyPI](https://pypi.org/project/spellbot/)"
        )
        author = "[@lexicalunit](https://github.com/lexicalunit)"
        embed.add_field(name="Author", value=author)
        embed.description = (
            "_A Discord bot for SpellTable._\n"
            "\n"
            "Use the command `!help` for usage details. Having issues with SpellBot? "
            "Please [report bugs](https://github.com/lexicalunit/spellbot/issues)!\n"
        )
        embed.url = "https://github.com/lexicalunit/spellbot"
        embed.set_footer(text="MIT © amy@lexicalunit et al")
        embed.color = discord.Color(0x5A3EFD)
        return embed, None

    @command(allow_dm=False)
    async def queue(self, channel, author, params):
        """
        Enter the queue for a game on SpellTable.
        """
        user = self.ensure_user_exists(author)
        if user.queue:
            await author.send(s("queue_already"), file=None)
            return

        group_size = 4
        user.enqueue(guild=channel.guild.id)
        group = Queue.playgroup(
            self.session, include=[user], guild=channel.guild.id, size=group_size
        )
        if not group:
            await author.send(s("queue"), file=None)
            return

        game_url = self.create_game()
        for user in group:
            discord_user = discord_user_from_id(channel, user.id)
            await discord_user.send(s("queue_ready", url=game_url))
        Queue.dequeue(self.session, group)


def get_db_url(database_env, fallback):  # pragma: no cover
    """Returns the database url from the environment or else the given fallback."""
    value = getenv(database_env, fallback)
    return value or fallback


@click.command()
@click.option(
    "-l",
    "--log-level",
    type=click.Choice(["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"]),
    default="ERROR",
)
@click.option("-v", "--verbose", count=True, help="Sets log level to DEBUG.")
@click.option(
    "-d",
    "--database-url",
    default=DEFAULT_DB_URL,
    help=(
        "Database url connection string; "
        "you can also set this via the SPELLBOT_DB_URL environment variable."
    ),
)
@click.option(
    "--database-env",
    default="SPELLBOT_DB_URL",
    help=(
        "By default SpellBot look in the environment variable SPELLBOT_DB_URL for the "
        "database connection string. If you need it to look in a different variable "
        "you can set it with this option. For example Heroku uses DATABASE_URL."
    ),
)
@click.version_option(version=__version__)
@click.option(
    "--dev",
    default=False,
    is_flag=True,
    help="Development mode, automatically reload bot when source changes",
)
def main(
    log_level, verbose, database_url, database_env, dev,
):  # pragma: no cover
    database_url = get_db_url(database_env, database_url)

    # We have to make sure that application directories exist
    # before we try to create we can run any of the migrations.
    ensure_application_directories_exist()

    client = SpellBot(
        token=getenv("SPELLBOT_TOKEN", None),
        auth=getenv("SPELLTABLE_AUTH", None),
        db_url=database_url,
        log_level="DEBUG" if verbose else log_level,
    )

    if dev:
        reloader = hupper.start_reloader("spellbot.main")
        reloader.watch_files(ASSET_FILES)

    client.run()


if __name__ == "__main__":
    main()



"""Initial revision

Revision ID: bb3268ba7108
Revises:
Create Date: 2020-06-24 18:02:42.616375

"""
import sqlalchemy as sa
from alembic import op

revision = "bb3268ba7108"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "authorized_channels",
        sa.Column("guild", sa.BigInteger(), nullable=False),
        sa.Column("name", sa.String(length=100), nullable=False),
        sa.PrimaryKeyConstraint("guild"),
    )
    op.create_table(
        "users",
        sa.Column("id", sa.BigInteger(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )


def downgrade():
    op.drop_table("users")
    op.drop_table("authorized_channels")



"""Queues

Revision ID: 6f160f1bf688
Revises: bb3268ba7108
Create Date: 2020-06-24 20:14:32.753166

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "6f160f1bf688"
down_revision = "bb3268ba7108"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "queue",
        sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
        sa.Column("user_id", sa.BigInteger(), nullable=False),
        sa.Column("guild", sa.BigInteger(), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(),
            nullable=False,
            server_default=sa.text("CURRENT_TIMESTAMP"),
        ),
        sa.PrimaryKeyConstraint("id"),
    )

    with op.batch_alter_table("queue") as b:
        b.create_foreign_key(
            "fk_queue_user", "users", ["user_id"], ["id"], ondelete="CASCADE"
        )


def downgrade():
    with op.batch_alter_table("queue") as b:
        b.drop_constraint("fk_queue_user", type_="foreignkey")

    op.drop_table("queue")



from os.path import dirname, realpath
from pathlib import Path
from string import Template

from yaml import load

try:
    from yaml import CLoader as Loader
except ImportError:  # pragma: no cover
    from yaml import Loader

PACKAGE_ROOT = Path(dirname(realpath(__file__)))
ASSETS_DIR = PACKAGE_ROOT / "assets"
STRINGS_DATA_FILE = ASSETS_DIR / "strings.yaml"
ASSET_FILES = [STRINGS_DATA_FILE]


def load_strings():
    with open(STRINGS_DATA_FILE) as f:
        return load(f, Loader=Loader)


def s(key, **kwargs):
    """Returns a string from strings.yaml asset with subsitutions."""
    if not s.strings:
        s.strings = load_strings()

    data = s.strings.get(key, "")
    assert data, f"error: missing strings key: {key}"
    return Template(data).substitute(kwargs)


s.strings = None



from os.path import dirname, realpath
from pathlib import Path

import alembic
import alembic.config
from sqlalchemy import (
    BigInteger,
    Column,
    DateTime,
    ForeignKey,
    Integer,
    String,
    and_,
    create_engine,
    func,
)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Session, relationship, sessionmaker

PACKAGE_ROOT = Path(dirname(realpath(__file__)))
ASSETS_DIR = PACKAGE_ROOT / "assets"
ALEMBIC_INI = ASSETS_DIR / "alembic.ini"
VERSIONS_DIR = PACKAGE_ROOT / "versions"


Base = declarative_base()


class AuthorizedChannel(Base):
    __tablename__ = "authorized_channels"
    id = Column(Integer, primary_key=True, nullable=False, autoincrement=True)
    guild = Column(BigInteger, primary_key=True, nullable=False)
    name = Column(String(100), nullable=False)


class User(Base):
    __tablename__ = "users"
    id = Column(BigInteger, primary_key=True, nullable=False)

    queue = relationship("Queue", uselist=False, back_populates="user")

    def enqueue(self, guild):
        session = Session.object_session(self)
        row = Queue(user_id=self.id, guild=guild)
        session.add(row)
        session.commit()
        return row


class Queue(Base):
    __tablename__ = "queue"
    id = Column(Integer, primary_key=True, nullable=False, autoincrement=True)
    user_id = Column(
        BigInteger, ForeignKey("users.id", ondelete="CASCADE"), nullable=False
    )
    guild = Column(BigInteger, primary_key=True, nullable=False)
    created_at = Column(DateTime, nullable=False, server_default=func.now())

    user = relationship("User", back_populates="queue")

    @classmethod
    def playgroup(cls, session, include, guild, size):
        others = (
            session.query(Queue)
            .filter(
                and_(Queue.guild == guild, ~Queue.user_id.in_([x.id for x in include]))
            )
            .order_by(Queue.created_at)
            .limit(size - len(include))
            .all()
        )
        if len(others) + len(include) < size:
            return None
        return (
            include
            + session.query(User).filter(User.id.in_(r.user_id for r in others)).all()
        )

    @classmethod
    def dequeue(cls, session, group):
        session.query(Queue).filter(
            Queue.user_id.in_([user.id for user in group])
        ).delete(synchronize_session=False)


def create_all(connection, db_url):
    config = alembic.config.Config(str(ALEMBIC_INI))
    config.set_main_option("script_location", str(VERSIONS_DIR))
    config.set_main_option("sqlalchemy.url", db_url)
    config.attributes["connection"] = connection
    alembic.command.upgrade(config, "head")


def reverse_all(connection, db_url):
    config = alembic.config.Config(str(ALEMBIC_INI))
    config.set_main_option("script_location", str(VERSIONS_DIR))
    config.set_main_option("sqlalchemy.url", db_url)
    config.attributes["connection"] = connection
    alembic.command.downgrade(config, "base")


class Data:
    """Persistent and in-memory store for user data."""

    def __init__(self, db_url):
        self.db_url = db_url
        self.engine = create_engine(db_url)
        self.conn = self.engine.connect()
        create_all(self.conn, db_url)
        self.Session = sessionmaker(bind=self.engine)
        self.metadata = Base.metadata



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['spellbot', 'spellbot.versions', 'spellbot.versions.versions']

package_data = \
{'': ['*'], 'spellbot': ['assets/*']}

install_requires = \
['alembic>=1.4.2,<2.0.0',
 'click>=7.1.2,<8.0.0',
 'discord-py>=1.3.3,<2.0.0',
 'dunamai>=1.2.0,<2.0.0',
 'hupper>=1.10.2,<2.0.0',
 'psycopg2-binary>=2.8.5,<3.0.0',
 'pydantic>=1.5.1,<2.0.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'pytz>=2020.1,<2021.0',
 'pyyaml>=5.3.1,<6.0.0',
 'requests>=2.24.0,<3.0.0',
 'sqlalchemy>=1.3.17,<2.0.0',
 'unidecode>=1.1.1,<2.0.0']

entry_points = \
{'console_scripts': ['spellbot = spellbot:main']}

setup_kwargs = {
    'name': 'spellbot',
    'version': '0.2.0',
    'description': 'A Discord bot for SpellTable',
    'long_description': '<img align="right" src="https://raw.githubusercontent.com/lexicalunit/spellbot/master/spellbot.png" />\n\n# SpellBot\n\n[![build][build-badge]][build]\n[![pypi][pypi-badge]][pypi]\n[![codecov][codecov-badge]][codecov]\n[![python][python-badge]][python]\n[![black][black-badge]][black]\n[![mit][mit-badge]][mit]\n\nA Discord bot for [SpellTable][spelltable].\n\n[![add-bot][add-img]][add-bot]\n\n## 📱 Using SpellBot\n\nOnce you\'ve connected the bot to your server, you can interact with it over\nDiscord via the following commands in any of the authorized channels.\n\n- `!about`: Get information about SpellBot\n- `!help`: Provides detailed help about all of the following commands\n- `!hello`: Says hello\n- `!queue`: Enter into a queue for an EDH game\n\n## WIP: Commands\n\n### Queue\n\nIn a text channel or as a DM to the bot:\n\n```text\n!queue <format> [power level]\n```\n\n- Verifies that you\'re not in a queue yet\n- Sends a DM to you indicating that you\'re in the queue\n- When matchmaking is complete:\n  - Creates a SpellTable link\n  - DMs everyone in the match the link\n\nQuestions:\n\n- What if someone want\'s to queue with someone else?\n- Maybe: `!queue <format> [power level] @with_this_person @and_this_person etc... ???\n\n### Status\n\nAs a DM to the bot:\n\n```text\n!status\n```\n\n- Gives you some information on your place in the queue\n- Possibly give you some status on your history of games and win/lose\n\n### Leaving\n\nAs a DM to the bot:\n\n```text\n!leave\n```\n\n- Removes you from the queue that you\'re in if you\'re in one\n\n### Reporting\n\nIn a text channel:\n\n```text\n!report win @username[, @username, @username, ...]\n!report draw @username[, @username, @username, ...]\n```\n\n- Automatically know what game based on the last game in which that user was a member of\n- Indicates that @username is the winner of their match\n- Potentially there could be multiple winners or draws\n- Allow for mistakes, user\'s should be able to run report multiple times\n- At some point after the first report, reporting needs be finalized\n- Do we need a `!report loss` command for any reason?\n\n### Moderation\n\nAs a DM to the bot by an authorized user:\n\n```text\n!ban @username <reason> [minutes]\n```\n\n- Block the user from being able to use the bot (optionally for the given minutes)\n\n```text\n!unban @username\n```\n\n- Remove user from the block list\n\n```text\n!bans\n```\n\n- Show the list of bans and reasons\n\n```text\n!ops @username\n```\n\n- Authorize someone else to be able to moderate\n\n```text\n!unops @username\n```\n\n- Remove someone from moderation\n\n```text\n!admin @username\n```\n\n- Authorize someone to be able to administrate\n\n```text\n!unadmin @username\n```\n\n- Remove someone from administrators\n\n```text\n!report ... ?\n```\n\n- Potentially have mods/admins be able to go in and manually change reports if need be\n\n### Concerns / Thoughts / Ideas\n\n- After a game is created, it needs to expire at some point\n- After a report is done, how long until the report command stops working\n- Does a user\'s game need to be reported before they can re-queue?\n- What happens if a game is never reported on?\n- When matchmaking w/ power levels, does it need to be a 100% match, +/- how much?\n- What about using ELO to match make?\n\n## 🤖 Running SpellBot\n\nFirst install `spellbot` using [`pip`](https://pip.pypa.io/en/stable/):\n\n```shell\npip install spellbot\n```\n\nProvide your Discord bot token with the environment variable `SPELLBOT_TOKEN`.\n\nBy default SpellBot will use sqlite3 as its database. You can however choose to\nuse another database by providing a [SQLAlchemy Connection URL][db-url]. This\ncan be done via the `--database-url` command line option or the environment\nvariable `SPELLBOT_DB_URL`. Note that, at the time of this writing, SpellBot is only\ntested against sqlite3 and PostgreSQL.\n\nMore usage help can be found by running `spellbot --help`.\n\n## 🐳 Docker Support\n\nYou can also run SpellBot via docker. See\n[our documentation on Docker Support](DOCKER.md) for help.\n\n## ❤️ Contributing\n\nIf you\'d like to become a part of the SpellBot development community please first\nknow that we have a documented [code of conduct](CODE_OF_CONDUCT.md) and then\nsee our [documentation on how to contribute](CONTRIBUTING.md) for details on\nhow to get started.\n\n---\n\n[MIT][mit] © [amy@lexicalunit][lexicalunit] et [al][contributors]\n\n[add-bot]:          https://discordapp.com/api/oauth2/authorize?client_id=725510263251402832&permissions=247872&scope=bot\n[add-img]:          https://user-images.githubusercontent.com/1903876/82262797-71745100-9916-11ea-8b65-b3f656115e4f.png\n[black-badge]:      https://img.shields.io/badge/code%20style-black-000000.svg\n[black]:            https://github.com/psf/black\n[build-badge]:      https://github.com/lexicalunit/spellbot/workflows/build/badge.svg\n[build]:            https://github.com/lexicalunit/spellbot/actions\n[codecov-badge]:    https://codecov.io/gh/lexicalunit/spellbot/branch/master/graph/badge.svg\n[codecov]:          https://codecov.io/gh/lexicalunit/spellbot\n[contributors]:     https://github.com/lexicalunit/spellbot/graphs/contributors\n[db-url]:           https://docs.sqlalchemy.org/en/latest/core/engines.html\n[lexicalunit]:      http://github.com/lexicalunit\n[mit-badge]:        https://img.shields.io/badge/License-MIT-yellow.svg\n[mit]:              https://opensource.org/licenses/MIT\n[pypi-badge]:       https://img.shields.io/pypi/v/spellbot\n[pypi]:             https://pypi.org/project/spellbot/\n[python-badge]:     https://img.shields.io/badge/python-3.7+-blue.svg\n[python]:           https://www.python.org/\n[spelltable]:       https://www.spelltable.com/\n',
    'author': 'Amy',
    'author_email': 'amy@lexicalunit.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/lexicalunit/spellbot',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4',
}


setup(**setup_kwargs)
